﻿using ProjectBattleship.Exceptions;
using ProjectBattleship.DrawingObject;
using System.Text;

namespace ProjectBattleship.CollectionGenericObjects;

public class StorageCollection<T>
    where T : DrawingWarship
{
    /// <summary>
    /// Словарь (хранилище) с коллекциями
    /// </summary>
    private Dictionary<CollectionInfo, ICollectionGenericObjects<T>> _storages;
    /// <summary>
    /// Возвращение списка названий коллекций
    /// </summary>
    public List<CollectionInfo> Keys => _storages.Keys.ToList();

    /// <summary>
    /// Ключевое слово, с которого должен начинаться файл
    /// </summary>
    private readonly string _collectionKey = "CollectionsStorage";

    /// <summary>
    /// Разделитель для записи ключа и значения элемента словаря
    /// </summary>
    private readonly string _separatorForKeyValue = "|";

    /// <summary>
    /// Разделитель для записей коллекции данных в файл
    /// </summary>
    private readonly string _separatorItems = ";";

    /// <summary>
    /// Конструктор
    /// </summary>
    public StorageCollection()
    {
        _storages = new Dictionary<CollectionInfo, ICollectionGenericObjects<T>>();
    }

    /// <summary>
    /// Добавление коллекции в хранилище
    /// </summary>
    /// <param name="collectionInfo">тип коллекции</param>
    public void AddCollection(CollectionInfo collectionInfo)
    {
        if (_storages.ContainsKey(collectionInfo)) throw new CollectionAlreadyExistsException(collectionInfo);
        if (collectionInfo.CollectionType == CollectionType.None)
            throw new CollectionTypeException("Пустой тип коллекции");
        if (collectionInfo.CollectionType == CollectionType.Massive)
            _storages[collectionInfo] = new MassiveGenericObjects<T>();
        else if (collectionInfo.CollectionType == CollectionType.List)
            _storages[collectionInfo] = new ListGenericObjects<T>();
    }

    /// <summary>
    /// Удаление коллекции
    /// </summary>
    /// <param name="collectionInfo">тип коллекции</param>
    public void DelCollection(CollectionInfo collectionInfo)
    {
        if (_storages.ContainsKey(collectionInfo))
            _storages.Remove(collectionInfo);
    }

    /// <summary>
    /// Доступ к коллекции
    /// </summary>
    /// <param name="collectionInfo"></param>
    /// <returns></returns>
    public ICollectionGenericObjects<T>? this[CollectionInfo collectionInfo]
    {
        get
        {
            if (_storages.ContainsKey(collectionInfo))
                return _storages[collectionInfo];
            return null;
        }
    }

    /// <summary>
    /// Сохранение информации по самолетам в файл
    /// </summary>
    /// <param name="filename"></param>
    /// <exception cref="EmptyFileExeption"></exception>
    public void SaveData(string filename)
    {
        if (_storages.Count == 0)
        {
            throw new EmptyFileException();
        }

        if (File.Exists(filename))
        {
            File.Delete(filename);
        }

        using (StreamWriter writer = new StreamWriter(filename))
        {
            writer.Write(_collectionKey);
            foreach (KeyValuePair<CollectionInfo, ICollectionGenericObjects<T>> value in _storages)
            {
                StringBuilder sb = new();
                sb.Append(Environment.NewLine);
                if (value.Value.Count == 0)
                {
                    continue;
                }

                sb.Append(value.Key);
                sb.Append(_separatorForKeyValue);
                sb.Append(value.Value.MaxCount);
                sb.Append(_separatorForKeyValue);
                foreach (T? item in value.Value.GetItems())
                {
                    string data = item?.GetDataForSave() ?? string.Empty;
                    if (string.IsNullOrEmpty(data))
                    {
                        continue;
                    }

                    sb.Append(data);
                    sb.Append(_separatorItems);
                }

                writer.Write(sb);
            }
        }
    }
    /// <summary>
    /// Загрузка информации по автомобилям в хранилище из файла
    /// </summary>
    /// <param name="filename">Путь и имя файла</param>
    /// <returns>true - загрузка прошла успешно, false - ошибка при загрузке данных</returns>
    public void LoadData(string filename)
    {
        if (!File.Exists(filename))
        {
            throw new Exceptions.FileNotFoundException(filename);
        }

        using (StreamReader fs = File.OpenText(filename))
        {
            string str = fs.ReadLine();
            if (string.IsNullOrEmpty(str))
            {
                throw new EmptyFileException(filename);
            }

            if (!str.StartsWith(_collectionKey))
            {
                throw new Exceptions.FileFormatException(filename);
            }

            _storages.Clear();
            string strs = "";
            while ((strs = fs.ReadLine()) != null)
            {
                string[] record = strs.Split(_separatorForKeyValue, StringSplitOptions.RemoveEmptyEntries);
                if (record.Length != 3)
                {
                    continue;
                }

                CollectionInfo? collectionInfo =
                    CollectionInfo.GetCollectionInfo(record[0]) ??
                    throw new CollectionInfoException("Не удалось определить информацию коллекции:" + record[0]);

                ICollectionGenericObjects<T>? collection =
                    StorageCollection<T>.CreateCollection(collectionInfo.CollectionType) ??
                    throw new CollectionTypeException("Не удалось определить тип коллекции:" + record[1]);
                collection.MaxCount = Convert.ToInt32(record[1]);

                string[] set = record[2].Split(_separatorItems, StringSplitOptions.RemoveEmptyEntries);
                foreach (string elem in set)
                {
                    if (elem?.CreateDrawingWarship() is T ship)
                    {
                        try
                        {
                            collection.Insert(ship);
                        }
                        catch (Exception ex)
                        {
                            throw new Exceptions.FileFormatException(filename, ex);
                        }
                    }
                }

                _storages.Add(collectionInfo, collection);
            }
        }
    }
    private static ICollectionGenericObjects<T>?
        CreateCollection(CollectionType collectionType)
    {
        return collectionType switch
        {
            CollectionType.Massive => new MassiveGenericObjects<T>(),
            CollectionType.List => new ListGenericObjects<T>(),
            _ => null,
        };
    }
}